#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <time.h>
#include "server.h"
#include "debug.h"

#define SERVER_BUFFER 32768
#define RETRY_DELAY 250000
#define CONNECTION_TIMEOUT 5

struct mpcs_conn
{
  time_t startup;
};

// todo locking schemes?

struct mpcs
{
  int port;
  char *address;
  char **cmdline;
  
  char *rdb;
  
  int pid;
  int w_sock;
  int r_sock;
} mpcs;

void chld_handler(int i)
{
  while(wait3(NULL, WNOHANG, (struct rusage * ) 0) > 0);

  mpcs.pid = 0;
  
  if(mpcs.r_sock != -1) close(mpcs.r_sock);
  if(mpcs.w_sock != -1) close(mpcs.w_sock);

  mpcs.r_sock = -1;
  mpcs.w_sock = -1;
}
    

int restart(struct mpcs *m)
{
  int r;

  int w_sock[2];
  int r_sock[2];
  
  if(m->pid)
  {
    kill(m->pid, SIGKILL);
    while(m->pid)
      usleep(RETRY_DELAY);
  }

  if(socketpair(AF_UNIX, SOCK_STREAM, 0, w_sock) < 0) return(0);
  if(socketpair(AF_UNIX, SOCK_STREAM, 0, r_sock) < 0) return(0);
     
  assert((r = fork()) != -1);

  
  if(r == 0)
  {
    dup2(w_sock[1], 0);
    dup2(r_sock[1], 1);
    dup2(r_sock[1], 2);
    if(execvp(m->cmdline[0], m->cmdline) == -1)
    {
      perror("execvp");
      _exit(0);
    }
  }
  
  m->r_sock = r_sock[0];
  m->w_sock = w_sock[0];
  m->pid = r;
  fprintf(stderr, "Successfully created a child process w/PID = %d, w_sock = %d\n", r, m->w_sock);
  return 1;
}

void * sconnect(void *udata, struct serv *s, struct conn *c)
{
  struct mpcs_conn *mc;
  
  assert((mc = (struct mpcs_conn*) malloc(sizeof(struct mpcs_conn))));
  mc->startup = time(NULL);
  return((void*) mc);
}

void sdisconnect(struct serv *s, struct conn *c, void *udata)
{
  free(udata);
}

int sinput(struct serv *s, struct conn *c, void *udata, char *buf, int len)
{
  if(!len) return 0;
  if(buf[len - 1] == '\r')
    buf[len - 1] = '\n';
    
  if(buf[len - 1] != '\n') return 0;
  
  if(mpcs.w_sock != -1)
  {
    if(write(mpcs.w_sock, buf, len) <= 0)
      perror("write");

    buf[len - 1] = 0;
    debug(">>> %s\n", buf);
  }

  return len;
}

int main(int argc, char **argv)
{
  struct serv * s;
  int opt;
  
  memset(&mpcs, 0, sizeof(struct mpcs));
  assert((mpcs.rdb = (char*) malloc(SERVER_BUFFER)));
  mpcs.port = 7400;
  
  while((opt = getopt(argc, argv, "p:a:h")) != EOF)
  {
    switch(opt)
    {
      case 'a': assert((mpcs.address = strdup(optarg))); break;
      case 'p':
        if(!(mpcs.port = atoi(optarg)))
        {
          fprintf(stderr, "Invalid port: %s\n", optarg);
          return 1;
        }
        break;
      case 'h': case '?':
        printf("mplayer control server (c) 2013 Mateusz 'mteg' Golicz\n");
        printf("Usage: %s [-p <RC port number>] [-a <RC address to bind to>] -- mplayer <mplayer parameters>\n", argv[0]);
        return 1;
    }
  }
  
  if(!argv[optind])
  {
    fprintf(stderr, "Incorrect usage, try %s -h for help!\n", argv[0]);
    return 1;
  }
  
  
  mpcs.cmdline = argv + optind;
  signal(SIGCHLD, chld_handler);
  
  if(!(s = server_init(mpcs.port, mpcs.address, 10, SERVER_BUFFER, &mpcs, sconnect, sdisconnect, sinput)))
  {
    fprintf(stderr, "Cannot initialize TCP server!\n");
    return 1;
  }
  
  while(1)
  {
    fd_set rfds, wfds, efds;
    int max = 0;
    struct conn *c;
    time_t now;
    
    if(!mpcs.pid)
    {
      usleep(RETRY_DELAY);
      restart(&mpcs);
    }
    
    /* Disconnect timed-out connections */
    now = time(NULL);
    for(c = s->connections; c; c = c->next)
    {
      struct mpcs_conn *mc = (struct mpcs_conn*) c->user_data;
      if(mc)
        if((now - mc->startup) >= CONNECTION_TIMEOUT)
          server_disconnect(c);
    }
    
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);

    if(mpcs.r_sock != -1)   
      FD_SET(max = mpcs.r_sock, &rfds);
    
    server_fdset(s, &rfds, &wfds, &efds, &max);
    if(select(max + 1, &rfds, &wfds, &efds, NULL) == -1)
    {
      perror("select");
      usleep(RETRY_DELAY);
    }
    server_perform(s);
    
    if(mpcs.r_sock != -1)
      if(FD_ISSET(mpcs.r_sock, &rfds))
      {
        int n = read(mpcs.r_sock, mpcs.rdb, SERVER_BUFFER - 1);
        if(n <= 0)
        {
          perror("read");
          usleep(RETRY_DELAY);
          restart(&mpcs);
        }
        else
        {
          mpcs.rdb[n] = 0;
          debug("<<< %s\n", mpcs.rdb);
          server_write(s, mpcs.rdb, n);
        }
      }
  }


  return 0;
}

