#define SERVER

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include "debug.h"
#include "server.h"

/* server_init - initialize the server */
struct serv * server_init(int port, char * address, int maxconn, int buffer_limit, void *user_data,
		void * (*connect)(void *, struct serv *, struct conn *),
		void (*disconnect)(struct serv *, struct conn *, void *),
		int (*input)(struct serv*, struct conn*, void*, char *, int))
{
	struct serv *s;
	struct sockaddr_in sin;
	int listeningSocket;
	
	/* Acquire socket */
	if((listeningSocket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1)
	{
		perror("socket");
		return(NULL);
	}
	
	/* Bind it to a port */
	sin.sin_family = PF_INET;
	sin.sin_port = htons(port);
	if(address)
	{
		if(!inet_aton(address, &sin.sin_addr))
		{
			perror("inet_aton()");
			return(NULL);
		}
	}
	else
		sin.sin_addr.s_addr = INADDR_ANY;
	
	if(bind(listeningSocket, (struct sockaddr*) &sin, sizeof(sin)) == -1)
	{
		perror("bind()");
		return(NULL);
	}
	
	/* Configure listening */
	if(listen(listeningSocket, 10) == -1)
	{
		perror("listen()");
		return(NULL);
	}

	/* Setup non-blocking operation */
	if(fcntl(listeningSocket, F_SETFL, O_NONBLOCK) == -1)
	{
		perror("fcntl(listeningSocket, F_SETFL, O_NONBLOCK)");
		return(NULL);
	}
	
	/* Initialize the server structure */
	assert(s = (struct serv*) malloc(sizeof(struct serv)));
	s->listening_socket = listeningSocket;
	s->conn_limit = maxconn;
	s->buff_limit = buffer_limit;

	assert(s->printf_buffer = (char*) malloc(s->buff_limit + 64));

	s->connections = NULL;
	s->conn_count = 0;
	s->user_data = user_data;
	
	s->connect = connect;
	s->disconnect = disconnect;
	s->input = input;
	
	return(s);
}

int server_perform(struct serv *s)
{
	struct sockaddr_in newSocketAddress;
	size_t addrlen;
	int sock;
	struct conn *c, *next;
	struct conn **prev;
	
	/* Loop through all connections and see if someone needs help ... */
	for(c = s->connections; c; c = c->next)
	{
		int we_read = 0;	/* Did we manage to get any new data? */
		
		if(c->o.size)	/* Does this connection have pending data to output? */
		{
			int wr_b;
			if((wr_b = write(c->fd, c->o.mem, c->o.size)) < 0)	/* Write data to socket */
			{
				if(errno != EAGAIN)	/* Write error, close the connection */
				{
					c->state = STATE_CLOSING;
					perror("write");
					debug("### Network ### [%d] Write error, closing connection to %s:%d\n", c->fd,
							inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
				}
			}
			else if(wr_b > 0)	/* Succesfully written some bytes */
			{
				if(wr_b < c->o.size)
					/* Maybe this is not the most elegant thing to do, but 
					   we expect that usually wr_b = c->o.size and the OS will take
					   care of further buffering. But if not - shift the buffer. */
					memmove(c->o.mem, ((unsigned char*) c->o.mem) + wr_b, c->o.size - wr_b);
				c->o.size -= wr_b;
			}
		}
		
		/* Now we'll try to see if there is new data available FROM this connection */
		while(1)
		{
			int rd_b;
			/* We'll be reading in chunks of maximally ALLOC_STEP bytes */
			
			/* So, if there is no more than ALLOC_STEP bytes free in the buffer, 
			   then we must allocate more memory... */
			if((c->i.alloc - c->i.size) <  ALLOC_STEP)
			{
				c->i.alloc += ALLOC_STEP;
				assert(c->i.mem = realloc(c->i.mem, c->i.alloc));
			}
			
			/* Haven't we exceeded the permitted buffer size? */
			if(c->i.alloc > s->buff_limit)
			{
				debug("### Network ### [%d] Read buffer exceeded, closing connection to %s:%d\n", c->fd,
						inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
				c->state = STATE_CLOSING;
				break;
			}
			
			/* Try reading to the free space in the buffer. */
			rd_b = read(c->fd, ((char*) c->i.mem) + c->i.size, ALLOC_STEP);
			if(rd_b < 0)
			{
				if(errno != EAGAIN)
				{
					c->state = STATE_CLOSING;
					perror("read");
					debug("### Network ### [%d] Read error, closing connection to %s:%d\n", c->fd,
							inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
				}
				/* Maybe there's simply no data available */
				break;
			}
			
			if(rd_b == 0)	/* Peer closed the connection */
			{
				debug("### Network ### [%d] EOF, closing connection to %s:%d\n", c->fd,
						inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
				c->state = STATE_CLOSING;
				break;
			}
			
			/* Getting here means rd_b > 0, so we have new data! */
			c->i.size += rd_b;
			
			we_read = 1;
			/* We'll try reading the buffer again (remember we're in a loop)! */
		}
		
		if(we_read)
		{
			if(c->i.size)
			{
				int read = s->input(s, c, c->user_data, c->i.mem, c->i.size);
				if(read < c->i.size)
					memmove(c->i.mem, c->i.mem + read, c->i.size - read);
				c->i.size -= read;
			}
		}		
	}

	/* Now we'll loop through the connections again and really CLOSE the ones that we want to close. */	
	prev = &s->connections;	/* Pointer to the place where we should UNLINK the current element */
	for(c = *prev; c; c = next)
	{
		next = c->next;	/* We must cache this because in a moment c->next might vanish */
		if(c->state == STATE_CLOSING)
		{
			debug("### Network ### [%d] Shutting down connection to %s:%d\n", c->fd,
							inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
			s->disconnect(s, c, c->user_data);
			close(c->fd);
			free(c->i.mem);
			free(c->o.mem);
			*prev = c->next;	/* Correct the next pointer in the previous element */
			free(c);
			s->conn_count --;	/* Decrease current connection count */
		}
		else
			prev = &c->next;	/* This connection is still valid,
						   but if the next one occurs to be scheduled for closing,
						   this is the place where it should be unlinked from */
	}
	
	/* Now check if there're any new connections */
	addrlen = sizeof(struct sockaddr_in);
	if((sock = accept(s->listening_socket, (struct sockaddr*) &newSocketAddress, (socklen_t*) &addrlen)) >= 0)
	{
		/* Set nonblock operation */
		if(fcntl(sock, F_SETFL, O_NONBLOCK) == -1)
		{
			perror("fcntl(newSocket, F_SETFL, O_NONBLOCK)");
			close(sock);
		}
		else if(s->conn_count > s->conn_limit)	/* See if we meet the limit of simultaneous connections */
			close(sock);
		else
		{
			struct conn * c;
			/* Allocate connection structure */
			assert(c = (struct conn*) malloc(sizeof(struct conn)));
			memset(c, 0, sizeof(struct conn));
			s->conn_count++;
			c->fd = sock;
			c->server = s;
			c->state = STATE_OPEN;	/* indicate that we're waiting for a request in this connection */
			c->user_data = s->connect(s->user_data, s, c);
			
			/* Copy the peer IP/port to the connection structure */
			memcpy(&c->peer, &newSocketAddress, sizeof(struct sockaddr_in));

			debug("### Network ### [%d] Accepted connection from %s:%d\n", sock,
						inet_ntoa(newSocketAddress.sin_addr), ntohs(newSocketAddress.sin_port));

			/* Link the connection to the connections list */
			c->next = s->connections;
			s->connections = c;
		}
	}
	else if(errno != EAGAIN)
	{
		perror("accept");
		return 1;	/* Trouble! */
	}
	return 0;
}

/* server_fdset - update fd_sets for select() */
void server_fdset(struct serv * s, fd_set *rd, fd_set *wr, fd_set *ex, int *maxfd)
{
	struct conn *c;
	
	/* Add the listening socket */
	FD_SET(s->listening_socket, rd);
	
	/* Check if we need to increase the highest fd number */
	if(s->listening_socket > *maxfd)
		*maxfd = s->listening_socket;

	/* Loop through all connections and add their sockets to the set */	
	for(c = s->connections; c; c = c->next)
	{
		if(c->fd > *maxfd)
			*maxfd = c->fd;
		FD_SET(c->fd, rd);	/* Add the socket to the read set, no matter what's happening */
		if(c->o.size)
			FD_SET(c->fd, wr);	/* Add the socket to the write set if there's some data pending
						   in the output buffer */
	}
}

/* server_write_conn() - append data to the connection output buffer */
void server_write_conn(struct serv *s, struct conn *c, void *ptr, size_t len)
{
	if(!s) s = c->server;
	/* Prepare more space if it will not fit... */
	while((c->o.size + len) > c->o.alloc)
	{
		c->o.alloc += ALLOC_STEP;
		debug("### Network ### Memory: Raising allocation in output buffer for %d to %d bytes\n", c->fd, c->o.alloc);
		assert(c->o.mem = realloc(c->o.mem, c->o.alloc));
		
		/* The peer does not manage to receive the data ... */
		if(c->o.alloc > s->buff_limit)
		{
			debug("### Network ### [%d] Write buffer exceeded, closing connection to %s:%d\n", c->fd,
					inet_ntoa(c->peer.sin_addr), ntohs(c->peer.sin_port));
			c->state = STATE_CLOSING;
		}
	}
	if(c->state != STATE_CLOSING)
	{
		/* Copy the data into the connection output buffer */
		memcpy(c->o.mem + c->o.size, ptr, len);
		c->o.size += len;
	}
}

void server_printf(struct serv *s, struct conn *c, char * fmt, ...)
{
	int len;
	va_list app;
	va_start(app, fmt);
	if(!s)
		s = c->server;
	len = vsnprintf(s->printf_buffer, s->buff_limit, fmt, app);
	if(len > 0)
		server_write_conn(s, c, s->printf_buffer, len);
}


/* server_write() - append data to output buffers of all connections */
void server_write(struct serv *s, void *ptr, size_t len)
{
	struct conn *c;
	
	for(c = s->connections; c; c = c->next)
		server_write_conn(s, c, ptr, len);
}

void server_disconnect(struct conn *c)
{
	c->state = STATE_CLOSING;
}

void server_emergency_shutdown(struct serv *s)
{
	struct conn *c;
	for(c = s->connections; c; c = c->next)
		close(c->fd);
	close(s->listening_socket);
}

